#!/usr/bin/env python

# 2010-10-12 13:58:15

###############################################################################
# Copyright (c) 2010, Vadim Shlyakhov
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#******************************************************************************

import os
import logging
from optparse import OptionParser
from subprocess import Popen, PIPE

knp_map={ # projection parameters
    'MERCATOR':
        {'PROJ': '+proj=merc',  'PP': '+lat_ts='},
    'POLYCONIC':
        {'PROJ': '+proj=poly',  'PP': '+lon_0='},
    'TRANSVERSE MERCATOR':
        {'PROJ': '+proj=tmerc', 'PP': '+lon_0=', 'P1': '+lat_0=', 'P2': '+k=', 'P3': '+y_0=', 'P4': '+x_0='},
    'UNIVERSAL TRANSVERSE MERCATOR':
        {'PROJ': '+proj=utm', 'PP': '+lon_0='},
    'GNOMONIC':
        {'PROJ': '+proj=gnom',  'PP': '+lon_0=', 'P1': '+lat_0='},
    #'LAMBERT CONFORMAL CONIC':  # not tested
    #    {'PROJ': '+proj=lcc', 
    }
knq_map={ # extra for BSB v. 3.xx
    'TRANSVERSE MERCATOR':
        {'P2': '+k=', 'P3': '+y_0=', 'P4': '+x_0='},
    }
datum_map={    
    'WGS84':                '+datum=WGS84',
    'NAD83':                '+datum=NAD83',
    'POTSDAM':              '+datum=potsdam',
    'EUROPEAN DATUM 1950' : '+towgs84=-84.0000,-97.0000,-117.0000 +ellps=intl',
    'EUROPEAN 1950':        '+towgs84=-84.0000,-97.0000,-117.0000 +ellps=intl',
    'ROMA DATUM 1940':      '+towgs84=-104.1,-49.1,-9.9,0.971,-2.917,0.714,-11.68 +ellps=intl',
    'ROMA 1940':            '+towgs84=-104.1,-49.1,-9.9,0.971,-2.917,0.714,-11.68 +ellps=intl',
    'HERMANSKOGEL DATUM':   '+datum=hermannskogel',
    'OSGB36':               '+towgs84=446.448,-125.157,542.060,0.1502,0.2470,0.8421,-20.4894 +ellps=airy',
    # 'LOCAL DATUM'
    # 'LOCAL DATUM UNKNOWN'
    }
guessed_datum_map={ # guess the datum by a comment/copyright string pattern
    'Croatia':
        # http://spatial-analyst.net/wiki/index.php?title=MGI_/_Balkans_coordinate_systems
        # '+ellps=bessel +towgs84=550.499,164.116,475.142,5.80967,2.07902,-11.62386,0.99999445824',
        # http://earth-info.nga.mil/GandG/coordsys/onlinedatum/DatumTable.html
        '+datum=hermannskogel',
    }
    
def l_d(smth):
    logging.debug(str(smth))
    
def command(params,stdin=None):
    l_d(' '.join(("'%s'" % i if ' ' in i else i for i in params)))
    process=Popen(params,stdin=PIPE, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    (stdout,stderr)=process.communicate(stdin)
    l_d((stdout,stderr))
    if process.returncode != 0: 
        raise Exception("*** External program failed: %s\n%s" % (' '.join(params),stderr))
    return stdout

def flatten(listOfLists): 
    out=[]
    for l in listOfLists: 
        out += list(l)
    return out

def hdr_parms(header, patt): 
    'filter header for params starting with "patt/", if knd is empty then return comment lines'
    if patt != '!': patt += '/'
    return [i[len(patt):] for i in header if i.startswith(patt)]

def hdr_parm2dict(header, knd):
    out={}
    for i in hdr_parms(header, knd)[0].split(','):
        if '=' in i:
            (key,val)=i.split('=')
            out[key]=val
        else:
            out[key] += ','+i
    return out

def hdr_read(kap_name): 
    'read KAP header'
    header=[]
    for l in open(kap_name, 'rU'):
        if '\x1A' in l:
            break
        if l.startswith((' ','\t')):
            header[-1] += ','+l.strip()
        else:
            header.append(l.strip())
    if not (header and header[0].startswith('!') and hdr_parms(header,'BSB') and hdr_parms(header,'KNP')): 
        raise Exception("*** invalid file: %s" % kap_name)
    l_d(header)
    return header

def srs_refs(bsb_header, options):
    'returns srs for the BSB chart projection and srs for the REF points'
    # Get a list of geo refs in tuples
    refs=map(eval,hdr_parms(bsb_header, 'REF'))
    l_d(refs)
    if options.srs:
        return(options.srs,refs)
    # get DTM northing, easting early
    dtm_parm=hdr_parms(bsb_header, 'DTM')[0] if not options.dtm else options.dtm
    dtm=eval(dtm_parm)
    l_d(dtm)
    # evaluate chart's projection
    knp_info=hdr_parm2dict(bsb_header, 'KNP')
    l_d(knp_info)
    proj=options.proj
    if not proj:
        try:
            parm_map=knp_map[knp_info['PR'].upper()]
        except KeyError: raise Exception('*** Unsupported projection')
        # get projection and parameters
        proj=[parm_map['PROJ']]
        if '+proj=utm' in proj[0]: 
            # GDAL 1.7.2 doesn't seem to make use of lon_0 with UTM, but BSB doesn't use zones
            #
            #zone=(eval(knp_info['PP']) + 3 % 360) // 6 + 30
            #proj.append('+zone='+str(zone))
            #if refs[0][3] < 0:
            #    proj.append('+south')
            northing='10000000' if refs[0][3] < 0 else '0' # Southern hemisphere?
            proj=['+proj=tmerc','+k=0.9996','+lon_0='+knp_info['PP'],'+x_0=500000','+y_0='+northing]
        else:
            proj+=[parm_map[i]+knp_info[i] for i in parm_map
                    if  i in knp_info and knp_info[i] not in ['NOT_APPLICABLE', 'UNKNOWN','']]
        try: # extra projection parameters for BSB 3.xx
            knq_info=hdr_parm2dict(bsb_header, 'KNQ')
            l_d(knq_info)
            parm_map=knq_map[knp_info['PR'].upper()]
            proj+=[parm_map[i]+knq_info[i] for i in parm_map
                    if  i in knq_info and knq_info[i] not in ['NOT_APPLICABLE', 'UNKNOWN','']]
        except: pass
        
    # evaluate chart's datum
    datum=options.datum
    if options.use_dtm or options.dtm:
        # Use DTM northing/easting correctiongs towards WGS84
        datum='+datum=WGS84'
        # alter refs as per dtm values
        refs=[ref[0:3]+(ref[3]+dtm[0]/3600,ref[4]+dtm[1]/3600) for ref in refs]
    if not datum:
        try:
            bsb_datum=knp_info['GD']
            datum=datum_map[bsb_datum.upper()]
        except KeyError: 
            # try to guess the datum by comment and copyright string(s)
            crr=" ".join(hdr_parms(bsb_header, '!')+hdr_parms(bsb_header, 'CRR'))
            try:
                datum=[guessed_datum_map[crr_patt] 
                    for crr_patt in guessed_datum_map if crr_patt in crr][0]
            except:
                if dtm == (0,0): 
                    datum='+datum=WGS84'
                else: # datum still not found
                    raise Exception('*** Unsupported or unknown datum. Try with "--use-dtm"')
            logging.warning('*** Unknown datum "%s", guessed as "%s"' % (bsb_datum,datum))
    srs=' '.join([' '.join(proj), datum,  '+nodefs'])
    l_d(srs)
    return (srs,refs)

def kap2vrt(kap):
    logging.info(kap)
    hdr=hdr_read(kap)           # Read BSB header
    
    bsb_info=hdr_parm2dict(hdr, 'BSB') # general BSB parameters
    kap_size=eval(bsb_info['RA'])

    (map_dir, map_fname)=os.path.split(kap)
    base=os.path.join(options.dest_dir, os.path.splitext(map_fname)[0])
    if options.long_names:
        base+=' - '+bsb_info['NA']
    aux_vrt= base+'.aux_vrt'    # auxilary VRT file
    out_vrt= base+'.vrt'        # output VRT file

    # Gather border polygon coordinates early
    plys=map(eval,hdr_parms(hdr, 'PLY'))
    if plys:
        # Create cutline
        latlong=''.join(['%f %f\n' % (i[2],i[1]) for i in plys])
        # convert cutline geo coordinates to pixel xy using GDAL's navive srs for this KAP
        poly_tr=command(['gdaltransform','-tps','-i','-t_srs','+proj=longlat', kap], latlong).splitlines()
        poly='POLYGON((%s))' % ','.join(poly_tr)
    if options.get_cutline: # print cutline then return
        try:
            print poly
            return
        except:
            raise Exception("*** No border polygon found")
        
    # Evaluate SRS
    (out_srs,refs)=srs_refs(hdr, options)
    # convert latlong gcps to projected coordinates
    latlong=''.join(['%f %f\n' % (ref[4],ref[3]) for ref in refs])
    refs_proj=[ i.split() for i in 
        command(['proj'] + out_srs.split(), latlong).splitlines()]
    l_d(refs_proj)
    # create vrt
    gcps=flatten([('-gcp', `i[0][1]`,`i[0][2]`,i[1][0],i[1][1]) for i in zip(refs, refs_proj)])
    cmd=['gdal_translate','-expand','rgb','-of','VRT',kap, out_vrt,'-a_srs',out_srs] + gcps
    if options.last_column_bug: # see http://trac.osgeo.org/gdal/ticket/3777
        cmd+=['-srcwin', '0', '0', str(kap_size[0]-1), str(kap_size[1]) ]
    command(cmd)

    if options.cut and plys: # polygon is there, then cut
        os.rename(out_vrt, aux_vrt)
        warp_cmd=['gdalwarp',
                  '-tps','-dstalpha',
                  '-wo','INIT_DEST=NO_DATA','-wo','CUTLINE='+poly,
                  '-of','VRT', aux_vrt, out_vrt]
        if options.blend_dist:
            warp_cmd=warp_cmd+['-wo','CUTLINE_BLEND_DIST=%s' % (options.blend_dist,)]
        command(warp_cmd)
    else:
        try:
            os.remove(aux_vrt)
        except: pass

if __name__=='__main__':
    usage = "usage: %prog [--cut] [--dest-dir=DEST_DIR] KAP_file..."
    parser = OptionParser(usage=usage,
        description="Extends builtin GDAL's support for BSB charts: WGS84 northing/easting, more projections, border line. "
        "The script converts .kap file with into GDAL .vrt, optionally clipping it out accroding to BSB border line.")
    parser.add_option("-d", "--debug", action="store_true", dest="debug")
    parser.add_option("-q", "--quiet", action="store_true", dest="quiet")
    parser.add_option("-t", "--dest-dir", default='',
        help='destination directory (default: current)')
    parser.add_option("-c", "--cut", action="store_true", 
        help='cut the raster as per BSB border line')
    parser.add_option("--cutline-blend", dest="blend_dist",default=None,
        help='CUTLINE_BLEND_DIST in pixels')
    parser.add_option("--get-cutline", action="store_true", 
        help='extract cutline polygon from KAP file then exit')
    parser.add_option("-u", "--use-dtm", action="store_true", 
        help='use BSB datum shifts record to convert to WGS84')
    parser.add_option("--dtm", dest="dtm",default=None,
        help='specify BSB datum shifts to convert SRS to WGS84')
    parser.add_option("--srs", default=None,
        help='override full chart with PROJ.4 definition of the spatial reference system')
    parser.add_option("--datum", default=None,
        help='override chart datum (PROJ.4 definition)')
    parser.add_option("--proj", default=None,
        help='override chart projection (PROJ.4 definition)')
    parser.add_option("-l", "--long-names", action="store_true", 
        help='give an output file a long name')
    parser.add_option("--last-column-bug", action="store_true", 
        help='some BSB files are missing value for last column, here is a workaround')

    (options, args) = parser.parse_args()
    
    if not args:
        parser.error('No input file(s) specified')
    logging.basicConfig(level=logging.DEBUG if options.debug else 
        (logging.ERROR if options.quiet else logging.INFO))
    l_d(options)

    if options.use_dtm:
        options.datum='+datum=WGS84'

    for f in args:
        kap2vrt(f)

